/*
 * $Header: /home/cvspublic/jakarta-slide/src/stores/org/apache/slide/store/txfile/TxXMLFileDescriptorsStore.java,v 1.3.2.2 2004/02/19 17:01:33 ozeigermann Exp $
 * $Revision: 1.3.2.2 $
 * $Date: 2004/02/19 17:01:33 $
 *
 * ====================================================================
 *
 * Copyright 1999-2002 The Apache Software Foundation 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package org.apache.slide.store.txfile;

import org.apache.slide.common.*;
import org.apache.slide.store.*;
import org.apache.slide.structure.*;

import org.apache.slide.security.*;
import org.apache.slide.lock.*;
import org.apache.slide.content.*;

import java.util.Enumeration;
import java.util.Hashtable;

/**
 * Transactional descriptors file store. 
 * Represents descriptor data as XML using {@link XMLResourceDescriptor}.
 *
 * <br><br>  
 * <em>Note</em>: To achieve search performance on properties comparable to a RDBMS some sort of index will be needed.  
 * 
 * @author <a href="mailto:ozeigermann@c1-fse.de">Oliver Zeigermann</a>
 * 
 */
public class TxXMLFileDescriptorsStore
    extends AbstractTxFileStoreService
    implements NodeStore, LockStore, RevisionDescriptorsStore, RevisionDescriptorStore, SecurityStore {

    protected static final String LOG_CHANNEL = TxFileContentStore.class.getName();
    
    protected static final String ENCODING_PARAMETER = "encoding";

    protected String characterEncoding = "UTF-8"; // a save choice

    public void setParameters(Hashtable parameters)
        throws ServiceParameterErrorException, ServiceParameterMissingException {

        super.setParameters(parameters);

        String encoding = (String) parameters.get(ENCODING_PARAMETER);
        if (encoding != null) {
            characterEncoding = encoding;
        }
    }

    /*
     * --- public methods of interface NodeStore ---
     *
     *  
     */

    public ObjectNode retrieveObject(Uri uri) throws ServiceAccessException, ObjectNotFoundException {
        XMLResourceDescriptor xfd = getFileDescriptor(uri);
        ObjectNode object = xfd.retrieveObject(uri);
        return object;
    }

    public void storeObject(Uri uri, ObjectNode object) throws ServiceAccessException, ObjectNotFoundException {
        XMLResourceDescriptor xfd = getFileDescriptor(uri);
        xfd.storeObject(uri, object);
    }

    public void createObject(Uri uri, ObjectNode object) throws ServiceAccessException, ObjectAlreadyExistsException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.create();
            xfd.storeObject(uri, object);
        } catch (ObjectNotFoundException e) {
            // should not happen, if it does, it is an error inside this store:
            throwInternalError("Newly created file vanished");
        }
    }

    public void removeObject(Uri uri, ObjectNode object) throws ServiceAccessException, ObjectNotFoundException {
        XMLResourceDescriptor xfd = getFileDescriptor(uri);
        xfd.delete();
    }

    /*
     * --- public methods of interface SecurityStore ---
     *
     *  
     */

    public void grantPermission(Uri uri, NodePermission permission) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.grantPermission(uri, permission);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public void revokePermission(Uri uri, NodePermission permission) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.revokePermission(uri, permission);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public void revokePermissions(Uri uri) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.revokePermissions(uri);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public Enumeration enumeratePermissions(Uri uri) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            return xfd.enumeratePermissions();
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
            return null; // XXX fake (is never called)
        }
    }

    /*
     * --- public methods of interface LockStore ---
     *
     *  
     */

    public void putLock(Uri uri, NodeLock lock) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.putLock(uri, lock);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public void renewLock(Uri uri, NodeLock lock) throws ServiceAccessException, LockTokenNotFoundException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.renewLock(uri, lock);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public void removeLock(Uri uri, NodeLock lock) throws ServiceAccessException, LockTokenNotFoundException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.removeLock(uri, lock);
        } catch (ObjectNotFoundException e) {
            throw new LockTokenNotFoundException(lock);
        }
    }

    public void killLock(Uri uri, NodeLock lock) throws ServiceAccessException, LockTokenNotFoundException {
        removeLock(uri, lock);
    }

    public Enumeration enumerateLocks(Uri uri) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            return xfd.enumerateLocks();
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
            return null; // XXX fake (is never called)
        }
    }

    /*
     * --- public methods of interface RevisionDescriptorsStore ---
     *
     *  
     */

    public NodeRevisionDescriptors retrieveRevisionDescriptors(Uri uri)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            return xfd.retrieveRevisionDescriptors(uri);
        } catch (ObjectNotFoundException e) {
            throw new RevisionDescriptorNotFoundException(uri.toString());
        }
    }

    public void createRevisionDescriptors(Uri uri, NodeRevisionDescriptors revisionDescriptors)
        throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.createRevisionDescriptors(uri, revisionDescriptors);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public void storeRevisionDescriptors(Uri uri, NodeRevisionDescriptors revisionDescriptors)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.storeRevisionDescriptors(uri, revisionDescriptors);
        } catch (ObjectNotFoundException e) {
            throw new RevisionDescriptorNotFoundException(uri.toString());
        }
    }

    public void removeRevisionDescriptors(Uri uri) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.removeRevisionDescriptors(uri);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    /*
     * --- public methods of interface RevisionDescriptorStore ---
     *
     *  
     */

    public NodeRevisionDescriptor retrieveRevisionDescriptor(Uri uri, NodeRevisionNumber revisionNumber)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            return xfd.retrieveRevisionDescriptor(uri, revisionNumber);
        } catch (ObjectNotFoundException e) {
            throw new RevisionDescriptorNotFoundException(uri.toString());
        }
    }

    public void createRevisionDescriptor(Uri uri, NodeRevisionDescriptor revisionDescriptor)
        throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.createRevisionDescriptor(uri, revisionDescriptor);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    public void storeRevisionDescriptor(Uri uri, NodeRevisionDescriptor revisionDescriptor)
        throws ServiceAccessException, RevisionDescriptorNotFoundException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.storeRevisionDescriptor(uri, revisionDescriptor);
        } catch (ObjectNotFoundException e) {
            throw new RevisionDescriptorNotFoundException(uri.toString());
        }
    }

    public void removeRevisionDescriptor(Uri uri, NodeRevisionNumber number) throws ServiceAccessException {
        try {
            XMLResourceDescriptor xfd = getFileDescriptor(uri);
            xfd.removeRevisionDescriptor(uri, number);
        } catch (ObjectNotFoundException e) {
            throwInternalError(e);
        }
    }

    /*
     * --- XMLFileDescriptor access and caching methods ---
     *
     *  
     */

    /**
     * Either returns a cached file descriptor or loads it from DB 
     */
    protected XMLResourceDescriptor getFileDescriptor(Uri uri) throws ServiceAccessException, ObjectNotFoundException {
        Object txId = getActiveTxId();
        XMLResourceDescriptor xfd = new XMLResourceDescriptor(uri, this, rm, txId, characterEncoding);
        return xfd;
    }
}